#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
#
# File:            froniusSolarAPI.py
#
# License:         Donationware, see attached LICENSE file for more
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.8a
#
# Date:            31.08.2023
#
# Description:     Library to communicate with the Fronius Solar API.
#
# Supported devices:
#                  Fronius Solar inverter devices.
#
# Requirements:    Python interpreter 2.7 or newer (www.python.org)
#
# Version history: ** 31.08.2023 v1.8a (Olli Lammi) **
#                  First version.
#
###########################################################################

# Imports

import string
import json
import urllib2

from modules.core import log


###########################################################################

# Constants

TYPE_METER = 'Meter'
TYPE_FLOW = 'Flow'

METER_URI = '/solar_api/v1/GetMeterRealtimeData.cgi?Scope=System'
FLOW_URI = '/solar_api/v1/GetPowerFlowRealtimeData.fcgi'


###########################################################################

# Classes

class FroniusSolarAPI(log.Logging):
    def __init__(self):
        log.Logging.__init__(self, 'FroniusSolarAPI')
        self.hostname = ''
        self.port = 80

    def runQueryCommand(self, cmds):
        res = {}
        data_meter = self.fetchData(TYPE_METER)
        data_flow = self.fetchData(TYPE_FLOW)

        for cmd in cmds:
            temp = string.split(cmd, '/')
            if temp[0] == TYPE_METER:
                if data_meter:
                    tempdata = data_meter['Body']['Data']
                    dev = temp[1]
                    key = temp[2]
                    if dev in tempdata:
                        if key in tempdata[dev]:
                            res[cmd] = str(tempdata[dev][key])
            elif temp[0] == TYPE_FLOW:
                if data_flow:
                    tempdata = data_flow['Body']['Data']
                    key1 = temp[1]
                    key2 = temp[2]
                    if key1 in tempdata:
                        if key2 in tempdata[key1]:
                            res[cmd] = str(tempdata[key1][key2])

        return res

    def initAPI(self):
        data = self.fetchData(TYPE_METER)
        if data:
            tempdata = data['Body']['Data']
            for dev in tempdata:
                manu = ''
                model = ''
                if 'Details' in tempdata[dev]:
                    manu = tempdata[dev]['Details']['Manufacturer']
                    model = tempdata[dev]['Details']['Model']
                self.Log("Detected Fronius device (id=%s): %s, %s" % (dev, manu, model))
            return True
        return False

    def fetchData(self, type = TYPE_METER):
        uri = ''
        if type == TYPE_METER:
          uri = METER_URI
        elif type == TYPE_FLOW:
          uri = FLOW_URI
        url = 'http://%s:%d%s' % (self.hostname, self.port, uri)
        self.Debug("Getting data using URL: %s" % url)

        data = None
        try:
            f = urllib2.urlopen(url)
            data = f.read()
            f.close()
        except:
            data = None

        if not data or len(data) <= 0:
            self.Debug("ERROR: Cannot get data.")
            return None

        try:
            data = json.loads(data)
            self.Debug("Retreived data: %s" % (repr(data)))
        except:
            self.Debug("ERROR: Error decoding JSON data from the received data.")
            return None

        return data
