#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
#
# File:            homeWizardAPI.py
#
# License:         Donationware, see attached LICENSE file for more
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.8d
#
# Date:            01.02.2024
#
# Description:     Library to communicate with the HomeWizard Energy API.
#
# Supported devices:
#                  HomeWizard Energy API enabled devices/services.
#
# Requirements:    Python interpreter 2.7 or newer (www.python.org)
#
# Version history: ** 30.01.2024 v1.8c (Olli Lammi) **
#                  First version.
#
#                  ** 01.02.2024 v1.8c (Olli Lammi) **
#                  Added error handling.
#
###########################################################################

# Imports

import string
import json
import urllib2

from modules.core import log


###########################################################################

# Constants


TYPE_BASIC = ''
TYPE_DATA = 'v1/data'


###########################################################################

# Classes

class HomeWizardAPI(log.Logging):
    def __init__(self):
        log.Logging.__init__(self, 'HomeWizardAPI')
        self.url = ''

    def runQueryCommand(self, cmds):
        res = {}
        data = self.fetchData(TYPE_DATA)

        uids = {}
        if data:
            if 'unique_id' in data:
                uids[data['unique_id']] = data
            if 'external' in data and data['external']:
                for item in data['external']:
                    if 'unique_id' in item:
                        uids[item['unique_id']] = item

        for cmd in cmds:
            temp = string.split(cmd, '/')
            if len(temp) != 2:
                self.Log("ERROR: HOMEWIZARD MEASURE point has to have exactly 2 fields separated by /. Skipping...")
                continue

            if not temp[0].strip() in uids:
                self.Log("ERROR: Unknown meter unique_id in MEASURE command: %s" % cmd)
                continue

            item = uids[temp[0].strip()]
            if not temp[1].strip() in item:
                self.Log("ERROR: Data value field not found in metered data: %s" % cmd)
                continue
            res[cmd] = str(item[temp[1].strip()])

        return res

    def initAPI(self):
        data = self.fetchData(TYPE_BASIC)
        if data:
            prodtype = ''
            prodname = ''
            fwversion = ''
            serial = ''
            if 'product_type' in data:
                prodtype = data['product_type']
            if 'product_name' in data:
                prodname = data['product_name']
            if 'firmware_version' in data:
                fwversion = data['firmware_version']
            if 'serial' in data:
                serial = data['serial']

            if prodtype:
                self.Log("Detected HomeWizard Energy API: %s %s %s %s" % (prodtype, prodname, fwversion, serial))
            else:
                self.Log("Could not detect HomeWizard Energy API.")
                return False

            data = self.fetchData(TYPE_DATA)
            if data:
                self.Log("Detected meters:")
                tempmodel = ''
                uid = ''
                if 'meter_model' in data:
                    tempmodel = data['meter_model']
                if 'unique_id' in data:
                    uid = data['unique_id']
                if uid:
                    self.Log("    meter: %s / %s" % (uid, tempmodel))
                if 'external' in data and data['external']:
                    for item in data['external']:
                        temptype = ''
                        uid = ''
                        if 'type' in item:
                            temptype = item['type']
                        if 'unique_id' in item:
                            uid = item['unique_id']
                        if uid:
                            self.Log("    meter: %s / %s" % (uid, temptype))

            return True
        return False

    def fetchData(self, type = TYPE_DATA):
        tempurl = self.url
        if tempurl[len(tempurl)-1] != '/':
            tempurl = tempurl + '/'
        tempurl = tempurl + type
        self.Debug("Getting data using URL: %s" % tempurl)

        data = None
        try:
            f = urllib2.urlopen(tempurl)
            data = f.read()
            f.close()
        except:
            data = None

        if not data or len(data) <= 0:
            self.Debug("ERROR: Cannot get data.")
            return None

        try:
            data = json.loads(data)
            self.Debug("Retreived data: %s" % (repr(data)))
        except:
            self.Debug("ERROR: Error decoding JSON data from the received data.")
            return None

        return data
