#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            husdataSerial.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.7l
#
# Date:            29.4.2020
#
# Description:     Library to communicate with the heat pumps using the
#                  Husdata.se H1-interface module.
#                  
# Supported devices: 
#                  Devices supported by Husdata.se H1 Interface. See
#                  document Husdata.se H1 Interface developers manual 
#                  (http://www.husdata.se/dl.asp?h=22344) for supported
#                  devices and available data point IDs. 
#
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
# 
#                  Python Serial Port Extension: pySerial  
#                  (http://pyserial.wiki.sourceforge.net/pySerial)
#
# Version history: ** 31.05.2016 v1.7h (Olli Lammi) **
#                  First version. 
#
#                  ** 22.04.2020 v1.7l (Olli Lammi) **
#                  Added optional XH-mode to Husdata data source module.
#
#                  ** 29.04.2020 v1.7l (Olli Lammi) **
#                  Added verbosity for testing and troubleshooting.
#                  Added H1-interface reset, when initializing.
#
#                  ** 30.04.2020 v1.7l (Olli Lammi) **
#                  Fixes to interface reset.
#
###########################################################################

# Imports

import sys, os 
import string
import time 

from modules.core import threads
from modules.core import log
from modules.utils import lockfile
from modules.utils import binary

# pySerial module, not loaded until init of HusdataSerial class instance
serial = None

###########################################################################

# Constants

BAUDRATE = 19200
TIMEOUT = 5


###########################################################################

# Data point types

TYPE_TEMP10 = 0
TYPE_ONOFF = 1
TYPE_INT10 = 2
TYPE_PERCENT10 = 3
TYPE_AMPERE10 = 4
TYPE_KWH10 = 5
TYPE_HOURS = 6
TYPE_MINS = 7
TYPE_DEGMINS = 8
TYPE_KW = 9
TYPE_CNT = 10


###########################################################################

# Classes

class HusdataSerial(threads.Thread, log.Logging):
    def __init__(self, port):
        threads.Thread.__init__(self)
        log.Logging.__init__(self, 'HusdataSerial')
        
        global serial
        serial = __import__('serial') 
        
        self.data = {}
        self.data_lock = threads.Lock()
  
        self.SERPORT = port
        self.DEVICEVERSION = ""
        self.serio = 0
        self.XH_MODE = False
          
        self.LOCK = None

    def setPort(self, port):
        self.SERPORT = port

    def setXHmode(self, xhmode):
        self.XH_MODE = xhmode

    def getDeviceVersion(self):
        return self.DEVICEVERSION
    
    def openPort(self):
        if not self.LOCK.lock():
            self.Log("ERROR: Unable to aquire lockfile for husdataSerial port: " + self.LOCK.getName())
            return 0

        try:
            self.serio = serial.Serial(port=self.SERPORT,baudrate=BAUDRATE, \
                                bytesize=serial.EIGHTBITS, \
                                parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, \
                                xonxoff=0, rtscts=0, dsrdtr=0, timeout=TIMEOUT)
            if not self.serio.isOpen():
                self.serio = 0
                self.LOCK.free()
                return 0 
        except:
            self.serio = 0
            self.LOCK.free()
            return 0
        
        self.Debug("Opened husdataSerial port: " + self.SERPORT)    
        
        if self.initPort():
            self.Debug("Initialized husdataSerial port: " + self.SERPORT)
            return 1    
        
        self.closePort()
        return 0
        
    def initPort(self):
        self.flushBuffer()

        # reset H1 interface
        i = 5
        while i > 0:
            self.Debug("Sending reset command to H1 interface.")
            (status, data) = sendCommand(self.serio, '!')
            if status == 1:
                if string.strip(data) == '!':
                    self.Log("Interface reset.")
                    break
                else:
                    self.Debug("Received non reset response message for reset command: " + data)
            i = i - 1
        if not i:
            self.Log("ERROR: Unable to reset Husdata H1 interface.")

        # get device version
        i = 5
        while i > 0:
            self.Debug("Sending version query to H1 interface.")
            (status, data) = sendCommand(self.serio, 'XV')
            if status == 1:
                if len(data) > 2 and string.upper(data)[:2] == 'XV':
                    self.DEVICEVERSION = data
                    break
                else:
                    self.Debug("Received non version response message for version query: " + data)
            i = i - 1
        if not i:
            self.Log("ERROR: Unable to get response to Husdata version query.")
        else:
            self.Debug("Husdata version " + self.DEVICEVERSION)
                
        self.start()
                
        return 1
        
    def flushBuffer(self):
        if self.serio:
            self.Debug("Flushing serial input buffer.")
            temp = ' '
            while len(temp) > 0 and self.serio.inWaiting() > 0:
                temp = self.serio.read(1)
        
    def closePort(self):
        if self.serio:
            if self.isRunning():
                self.terminate()
                self.wait()
            try:
                self.flushBuffer()
            except:
                pass
            try:
                self.serio.close()
            except:
                pass
            self.serio = 0
        self.DEVICEVERSION = ''
        self.LOCK.free()
        self.Debug("Closed husdataSerial port: " + self.SERPORT)

    def isOpen(self):
        if self.serio and self.serio.isOpen():
            return 1
        return 0

    def run(self):
        res = ""
        errcnt = 0
        initlevel = 3
        while self.isRunning():
            try:
                blocktime = time.time()
                temp = self.serio.read()
                if len(temp) <= 0 and time.time() - blocktime < (TIMEOUT / 2.0):
                    errcnt = errcnt + 1
                    if errcnt > 5:
                        self.Log("Repeating error while reading Husdata serial data. Terminating thread.")
                        self.setFail()
                elif len(temp) <= 0:
                    if initlevel > 0 and len(res) <= 0:
                        if initlevel == 3:
                            self.Debug("Requesting registers (XR).")
                            self.serio.write('XR\x0D')
                            self.serio.flush()    
                        elif initlevel == 2:
                            self.Debug("Requesting heat pump registers (XS).")
                            self.serio.write('XS\x0D')
                            self.serio.flush()    
                        elif initlevel == 1:
                            if self.XH_MODE:
                                self.Debug("Enabling XH-mode.")
                                self.serio.write('XH\x0D')
                                self.serio.flush()    
                        initlevel = initlevel - 1

                if len(temp) > 0:
                    errcnt = 0
                    res = res + temp
                stat = 1
                while stat == 1 and len(res) > 0:
                    (stat, msg, res) = checkMessage(res)
                    if stat == 1 and len(msg) > 0:
                        self.Debug("Received Husdata serial data: " + repr(msg))
                        self.handleRow(msg)
            except Exception, e:
                self.Log("Exception: " + e.__str__())
                self.setFail()
            except IOError, ioe:
                self.Log("IOError: " + ioe.__str__())
                self.setFail()

        self.closePort()

    def handleRow(self, row):
        row = string.upper(string.split(string.strip(row), ' ')[0])
        if len(row) <= 2:
            return
        
        if not row.startswith("XR"):
            return
        
        row = row[2:]
        if len(row) != 8:
            return
        
        self.Debug("Received Husdata register data: " + row)
        dtype = int(row[0], 16)
        did = int(row[1:4], 16)
        value = int(row[4:], 16)

        if self.data_lock.lock_wait():
            if not self.data.has_key(did):
                self.Log("Received new Husdata data point ID: " + row[1:4])
            self.data[did] = convertHusdataMessage(dtype, value)
            self.Debug("Received value for data point ID: %s, value: %s" % (row[1:4], self.data[did]))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")

    def startModule(self):
        if self.LOCK == None:
            self.LOCK = lockfile.LockFile(self.SERPORT)
            
        if not self.isOpen():
            self.openPort()
        if self.isOpen():
            self.start()
            return 1
        else:
            self.setFail()
            self.setTerminated()
            
        return 0

    def runQueryCommand(self, cmd):
        if self.hasTerminated() and self.isFailed():
            self.startModule()

        try:
            did = int(cmd, 16)
        except:
            self.Log("ERROR: Invalid Husdata data point key: " + cmd)
            return ""
            
        res = self.runQueryId(did)
        if res != None:
            return res
        return ""
    
    def runQueryId(self, id):
        self.Debug("Running husdataSerial query with id %0.3X" % id)

        res = None
        if self.data_lock.lock_wait():        
            if self.data.has_key(id):
                res = self.data[id]
            else:
                self.Debug("No data found with id " + repr(id))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")
        return res   

###########################################################################

# Functions

# returns (status, response)
# status: 1 = ok, -1 = timeout, 0 = error in response
def sendCommand(serio, cmd):
    serio.write(cmd + '\x0D')
    serio.flush()

    res = ""
    temp = " "
    while len(temp) > 0:
        temp = serio.read()
        if len(temp) <= 0:
            if len(res) > 0:
                (stat, data, res) = checkMessage(res) 
                return (stat, data)
            else:
                return (-1, '')
        res = res + temp
        (stat, data, res) = checkMessage(res)
        if stat == 0 or stat == 1:
            return (stat, data)

def checkMessage(res):
    for i in range(len(res)):
        if res[i] == '\x0D' or res[i] == '\x0A' or res[i] == '\x00':
            return (1, res[:i], res[i+1:])
    return (-1, '', res)

    
def convertHusdataMessage(type, data):
    temp = data
    if type in (TYPE_TEMP10, TYPE_INT10, TYPE_PERCENT10, TYPE_AMPERE10, TYPE_KWH10) :
        temp = binary.twosComplementToInt(temp, 16)
        temp = "%.1f" % (temp / 10.0)
        return temp
    elif type == TYPE_ONOFF:
        temp = "%d" % temp
        return temp
    elif type in (TYPE_HOURS, TYPE_MINS, TYPE_DEGMINS, TYPE_KW, TYPE_CNT):
        temp = binary.twosComplementToInt(temp, 16)
        temp = "%d" % (temp, )
        return temp
    else:
        temp = "%d" % (temp, )
        return temp

    return ""
