#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            stEltronCanSerial.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         0.1a
#
# Date:            18.11.2015
#
# Description:     Library to communicate with the Stiebel Eltron Can bus 
#                  type heat pump controllers.
#                  
# Supported devices: 
#                  Stiebel Eltron devices with can bus. 
#                  Tested with:
#                      DEFAULT: xxx
#
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
# 
#                  Python Serial Port Extension: pySerial  
#                  (http://pyserial.wiki.sourceforge.net/pySerial)
#
# Version history: ** 18.11.2015 v0.1a (Olli Lammi) **
#                  First test version. No datapoints. 
#
###########################################################################

# Imports

import sys, os 
import string, math
import tempfile
import time 

from modules.core import threads
from modules.core import log
from modules.utils import lockfile
from modules.utils import binary
from modules.utils import can232Serial


###########################################################################

# Constants

DATAVALID = 240

###########################################################################

# Device configuration data

TYPE_UINT8 = 1
TYPE_UINT16 = 2

STELTRONCAN_DEVICES = { 'DEFAULT': [ 20, [ \
                               ] ] \
                  }

   

###########################################################################

# Classes

class StiebelEltronCanSerial(log.Logging, can232Serial.CANFrameListener):
    def __init__(self, port, device, baudrate):
        log.Logging.__init__(self, 'StiebelEltronCanSerial')
               
        self.data = {}
        self.data_lock = threads.Lock()
  
        self.SERPORT = port
        self.STELTRONCAN_DEVICE = device
        self.BAUDRATE = baudrate
        self.can232 = None
          
    def setPort(self, port):
        self.SERPORT = port
        
    def setDevice(self, device):
        self.STELTRONCAN_DEVICE = device
        
    def setBaudrate(self, baudrate):
        self.BAUDRATE = baudrate      
                    
    def frameReceived(self, id, data):
        if self.data_lock.lock_wait():
            self.data[id] = [time.time(), data]
            self.Debug("Got data for id %d/0x%X, data lenght %d: %s" % (id, id, len(data), repr(data)))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")

    def startModule(self):
        if not self.isOpen():
            self.openPort()
        if self.isOpen():
            return 1

        return 0
    
    def terminate(self):
        if self.isOpen():
            self.closePort()
        return 1
                   
    def openPort(self):
        if not STELTRONCAN_DEVICES.has_key(self.STELTRONCAN_DEVICE):
            self.Log("ERROR: Invalid STELTRONCAN device type: " + self.STELTRONCAN_DEVICE)
            return 0
        canspeedkbps = STELTRONCAN_DEVICES[self.STELTRONCAN_DEVICE][0]

        self.can232 = can232Serial.Can232Serial(self.SERPORT, canspeedkbps, self.BAUDRATE)
        self.can232.setFrameListener(self)
        if not self.can232.openPort():
            self.Debug("Cannot open stEltronCanSerial port: " + self.SERPORT)
            self.can232 = None
            return 0
        
        self.Debug("Opened stEltronCanSerial port: " + self.SERPORT)
        return 1
        
    def closePort(self):
        if self.can232:
            try:
                self.can232.closePort()
            except:
                pass
            self.can232 = None
        self.Debug("Closed stEltronCanSerial port: " + self.SERPORT)

    def isOpen(self):
        if self.can232 and self.can232.isOpen():
            return 1
        return 0

    def getCanDeviceVersion(self):
        if self.can232:
            return self.can232.getDeviceVersion()
        return ""

    def runQueryCommand(self, cmd):
        if not self.isOpen():
            if not self.startModule():
                self.Log("Could not restart module after being failed.")
            else:
                self.Log("Module restarted after being failed.")
        
        id = 0
        type = 0
        for t in STELTRONCAN_DEVICES[self.STELTRONCAN_DEVICE][1]:
            if t[1] == cmd:
                id = t[0]
                type = t[2]
                break

        if type == 0:
            self.Log("ERROR: Invalid STELTRONCAN key: " + cmd)
            return ""

        res = self.runQueryId(id)
        if res != None:
            return convertStEltronCanMessage(type, res)
        return ""

    def runQueryId(self, id):
        self.Debug("Running stEltronCanSerial query with id %d/0x%X" % (id, id))

        res = None
        if self.data_lock.lock_wait():
            if self.data.has_key(id):
                temp = self.data[id]
                if (time.time() - temp[0]) <= DATAVALID:
                    res = temp[1]
                else:
                    self.Debug("Data expired with id %d/0x%X" % (id, id))
            else:
                self.Debug("No data found with id %d/0x%X" % (id, id))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")
        return res   

        
###########################################################################

# Functions
    
def convertStEltronCanMessage(type, data):
    if type == TYPE_UINT8:
        temp = ord(data[0])
        temp = "%d" % temp
        return temp
    elif type == TYPE_UINT16:
        temp = (ord(data[0]) << 8) + ord(data[1])
        temp = "%d" % temp
        return temp

    return ""

